"""Define the deconstruction process.

"""


from collections import deque
from multiprocessing import Queue, Event

from cml.shared.logger import DeconstInfo
from cml.domain.knowledge import TS_QUEUE
from cml.domain.deconstruction.z_sigma import ZSigmaDeconstructor
from cml.domain.deconstruction.complete import CompleteDeconstructor
from cml.domain.deconstruction.t_sigma import (
    TSigmaDeconstructor,
    Start,
    Apply,
    Upgrade,
    Clean
)


__all__ = (
    "Deconstructor",
)


class AbstractDeconstructor:

    def __init__(self, mode, settings, reconstructor, learnblock_constructor,
                 relative_finder, cluster_finder):
        self.mode = mode
        self.settings = settings
        self.reconstructor = reconstructor
        self.learnblock_constructor = learnblock_constructor
        self.relative_finder = relative_finder
        self.cluster_finder = cluster_finder

    def __str__(self):
        return "\n".join([
                "{:20}: {}".format("Mode", str(self.mode)),
                str(self.settings),
                str(self.reconstructor)
            ]
        )

    def __repr__(self):
        return ", ".join([
                "{}={}".format("mode", repr(self.mode)),
                "{}".format(repr(self.settings)),
                "{}".format(repr(self.reconstructor))
            ]
        )

    def deconstruct(self, *args, **kwargs):
        raise NotImplementedError

    def to_dict(self):
        return {
            "mode": self.mode,
            "settings": self.settings,
            "reconstructor": self.reconstructor,
            "learnblock_constructor": self.learnblock_constructor,
            "relative_finder": self.relative_finder,
            "cluster_finder": self.cluster_finder
        }

    def find_relatives(self, pragmatic):

        complete_relatives = self.relative_finder.find(
            tuple(["complete", ]), pragmatic,
            self.settings.deconst_full_tolerance)

        t_sigma_relatives = self.relative_finder.find(
            tuple(["T", "Sigma"]), pragmatic,
            self.settings.deconst_full_tolerance)

        sigma_z_relatives = self.relative_finder.find(
            tuple(["Sigma", "Z"]), pragmatic,
            self.settings.deconst_full_tolerance)

        yield complete_relatives
        yield t_sigma_relatives
        yield sigma_z_relatives


class SerialDeconstructor(AbstractDeconstructor):

    def __str__(self):
        return super().__str__()

    def __repr__(self):
        return super().__repr__()

    def deconstruct(self, level, pragmatic, database):
        relatives = iter(super().find_relatives(pragmatic))

        yield self.run_complete_deconstruction(
            level, pragmatic, database, next(relatives)
        )

        yield self.run_t_sigma_deconstruction(
            level, pragmatic, database, next(relatives)
        )

        yield self.run_sigma_z_deconstruction(
            level, pragmatic, database, next(relatives)
        )

    def run_complete_deconstruction(self, level, pragmatic, database, relatives):
        result_deque = deque()
        CompleteDeconstructor(
            level, database, self.settings, self.reconstructor, pragmatic,
            relatives, result_deque, cluster_finder=self.cluster_finder
        ).run()

        for log in result_deque:
            self.clean_ts_queue(log.deleted)
            yield log

    def run_t_sigma_deconstruction(self, level, pragmatic, database, relatives):
        result_deque = deque()
        TSigmaDeconstructor(
            self.learnblock_constructor, database, self.settings
        ).run_serial(level, relatives, pragmatic, result_deque)

        for log, level_, learnblock in result_deque:
            TS_QUEUE.append((level_, learnblock))
            yield log

    def run_sigma_z_deconstruction(self, level, pragmatic, database, relatives):
        result_deque = deque()
        ZSigmaDeconstructor(
            level, database, self.settings, self.reconstructor, pragmatic,
            relatives, result_deque
        ).run()

        for log in result_deque:
            self.clean_ts_queue(log.deleted)
            yield log

    def clean_ts_queue(self, deleted):
        queue_copy = TS_QUEUE.copy()
        for level, lb in queue_copy:
            for model in deleted:
                if model in lb.origin:
                    TS_QUEUE.remove((level, lb))


class ParallelDeconstructor(AbstractDeconstructor):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Used if parallelization is on
        self.tsigma_process = None
        self.tsigma_inputs = None
        self.tsigma_outputs = None
        self.upgrade_event = None
        self.apply_event = None
        self.clear_event = None

    def __str__(self):
        return super().__str__()

    def __repr__(self):
        return super().__repr__()

    def deconstruct(self, level, pragmatic, database):
        self.start_background_tsigma(database)
        relatives = iter(self.find_relatives(pragmatic))
        complete_relatives = next(relatives)
        t_sigma_relavites = next(relatives)
        self.tsigma_inputs.put(Start(level=level, relatives=t_sigma_relavites,
                                     pragmatic=pragmatic))

        yield self.run_complete_deconstruction(
            level, pragmatic, database, complete_relatives
        )

        yield self.run_t_sigma_deconstruction()

        yield self.run_sigma_z_deconstruction(
            level, pragmatic, database, next(relatives)
        )

        self.upgrade_event.clear()
        self.apply_event.clear()
        self.clear_event.clear()

    def start_background_tsigma(self, database):
        if not self.tsigma_process:
            self.tsigma_inputs = Queue()
            self.tsigma_outputs = Queue()
            self.upgrade_event = Event()
            self.apply_event = Event()
            self.clear_event = Event()
            self.tsigma_process = TSigmaDeconstructor(
                self.learnblock_constructor, database, self.settings,
                self.tsigma_inputs, self.tsigma_outputs, self.upgrade_event,
                self.apply_event, self.clear_event)

            self.tsigma_process.daemon = True
            self.tsigma_process.start()

    def run_complete_deconstruction(self, level, pragmatic, database, relatives):
        result_deque = deque()
        CompleteDeconstructor(
            level, database, self.settings, self.reconstructor, pragmatic,
            relatives, result_deque, cluster_finder=self.cluster_finder
        ).run()

        while result_deque:
            log_info = result_deque.pop()

            if log_info.status and self.settings.deconst_mode == "minimal":
                self.tsigma_inputs.put(Clean())
                self.clear_event.wait()

            yield log_info

            self.tsigma_inputs.put(Upgrade(deletes=log_info.deleted))

    def run_t_sigma_deconstruction(self):
        self.tsigma_inputs.put(Apply())
        self.apply_event.wait()

        while not self.tsigma_outputs.empty():
            log_info, level, lb = self.tsigma_outputs.get(timeout=0)
            TS_QUEUE.append((level, lb))
            yield log_info

    def run_sigma_z_deconstruction(self, level, pragmatic, database, relatives):
        result_deque = deque()
        ZSigmaDeconstructor(
            level, database, self.settings, self.reconstructor, pragmatic,
            relatives, result_deque
        ).run()

        while result_deque:
            log_info = result_deque.pop()
            yield log_info


class Deconstructor:
    def __init__(self, mode, settings, reconstructor, learnblock_constructor,
                 relative_finder, cluster_finder):

        self.state = SerialDeconstructor(
            mode, settings, reconstructor, learnblock_constructor,
            relative_finder, cluster_finder)

    def __str__(self):
        return str(self.state)

    def __repr__(self):
        return repr(self.state)

    def transit(self, parallel):
        if parallel:
            self.state = ParallelDeconstructor(**self.state.to_dict())
        else:
            self.state = SerialDeconstructor(**self.state.to_dict())

    def deconstruct(self, level, pragmatic, database):
        changes_done_status = []
        for decon_type in self.state.deconstruct(level, pragmatic, database):
            for log in decon_type:
                changes_done_status.append(log.status)
                yield log
                # First deconstruction successfully ended and deconst mode
                # says, end the deconstruction.
                if log.status and self.deconst_mode == "minimal":
                    return

                # The new reconstructed pragmatic model was deleted, so end the
                # whole deconstruction process.
                elif pragmatic in log.deleted:
                    return

        if not changes_done_status or not all(changes_done_status):
            database.insert(pragmatic)
            yield DeconstInfo(pragmatic=pragmatic, inserted=[pragmatic])

    @property
    def mode(self):
        return self.state.mode

    @property
    def relative_finder(self):
        return self.state.relative_finder

    @property
    def deconst_mode(self):
        return self.state.settings.deconst_mode

    @deconst_mode.setter
    def deconst_mode(self, value):
        self.state.settings.deconst_mode = value

    @property
    def deconst_strategy(self):
        return self.state.settings.deconst_strategy

    @deconst_strategy.setter
    def deconst_strategy(self, value):
        self.state.settings.deconst_strategy = value

    @property
    def deconst_max_distance_t(self):
        return self.state.settings.deconst_max_distance_t

    @deconst_max_distance_t.setter
    def deconst_max_distance_t(self, value):
        self.state.settings.deconst_max_distance_t = value

    @property
    def deconst_full_tolerance(self):
        return self.state.settings.deconst_full_tolerance

    @deconst_full_tolerance.setter
    def deconst_full_tolerance(self, value):
        self.state.settings.deconst_full_tolerance = value

    @property
    def force_time_expansion(self):
        return self.state.settings.force_time_expansion

    @force_time_expansion.setter
    def force_time_expansion(self, value):
        self.state.settings.force_time_expansion = value

    @property
    def allow_weak_reliability(self):
        return self.state.settings.allow_weak_reliability

    @allow_weak_reliability.setter
    def allow_weak_reliability(self, value):
        self.state.settings.allow_weak_reliability = value

    @property
    def learn_block_minimum(self):
        return self.state.settings.learn_block_minimum

    @learn_block_minimum.setter
    def learn_block_minimum(self, value):
        self.state.settings.learn_block_minimum = value

    @property
    def min_reliability(self):
        return self.state.settings.min_reliability

    @min_reliability.setter
    def min_reliability(self, value):
        self.state.settings.min_reliability = value


# class Deconstructor2:
#     TIME_COLUMN = "T"
#     SUBJECT_COLUMN = "Sigma"
#     PURPOSE_COLUMN = "Z"
#
#     def __init__(self, mode, relative_finder, cluster_finder,
#                  build_new_learnblock, settings, reconstructor):
#         self.mode = mode
#         self.settings = settings
#         self.build_new_learnblock = build_new_learnblock
#         self.relative_finder = relative_finder
#         self.cluster_finder = cluster_finder
#         self.reconstructor = reconstructor
#
#         # Used if parallelization is on
#         self.tsigma_process = None
#         self.tsigma_inputs = None
#         self.tsigma_outputs = None
#         self.upgrade_event = None
#         self.apply_event = None
#         self.clear_event = None
#
#     @reset_resources
#     def deconstruct_parallel(self, level, pragmatic, block, database):
#
#
#         decon_one = self.start_background_tsigma(database)
#         relatives = self.relative_finder.find(("T", "Sigma"),
#             pragmatic, self.settings.deconst_full_tolerance)
#         self.tsigma_inputs.put(Start(level=level,
#                                      relatives=relatives,
#                                      pragmatic=pragmatic,
#                                      block=block))
#
#         # complete deconstruction
#         deq_two = deque()
#         relatives = list(self.relative_finder.find(tuple(("complete", )),
#             pragmatic, self.settings.deconst_full_tolerance))
#         decon_two = CompleteDeconstructor(level,
#                                           database,
#                                           self.settings,
#                                           self.reconstructor,
#                                           pragmatic,
#                                           relatives,
#                                           block,
#                                           deq_two,
#                                           cluster_finder=self.cluster_finder)
#         decon_two.run()
#
#         while deq_two:
#             log_info = deq_two.pop()
#             yield log_info
#             info_status.append(log_info.status)
#             if log_info.status and self.settings.deconst_mode == "minimal":
#                 self.tsigma_inputs.put(Clean())
#                 self.clear_event.wait()
#                 return
#
#             self.tsigma_inputs.put(Upgrade(deletes=log_info.deleted))
#
#         self.tsigma_inputs.put(Apply())
#         self.apply_event.wait()
#         while not self.tsigma_outputs.empty():
#             log_info, level, lb = self.tsigma_outputs.get(timeout=0)
#             TS_QUEUE.append((level, lb))
#             yield log_info
#             info_status.append(log_info)
#             if log_info.status and self.settings.deconst_mode == "minimal":
#                 return
#
#         deq_three = deque()
#         relatives = list(self.relative_finder.find(("Sigma", "Z"),
#             pragmatic, self.settings.deconst_full_tolerance))
#         decon_three = ZSigmaDeconstructor(level,
#                                           database,
#                                           self.settings,
#                                           self.reconstructor,
#                                           pragmatic,
#                                           relatives,
#                                           block,
#                                           deq_three)
#         decon_three.run()
#         while deq_three:
#             log_info = deq_three.pop()
#             yield log_info
#             info_status.append(log_info)
#             if log_info.status and self.settings.deconst_mode == "minimal":
#                 return
#
#         if not info_status or not all(info_status):
#             database.insert(pragmatic)
#             yield self.arrange_log(pragmatic)
#
#     def start_background_tsigma(self, database):
#         if not self.tsigma_process:
#             self.tsigma_inputs = Queue()
#             self.tsigma_outputs = Queue()
#             self.upgrade_event = Event()
#             self.apply_event = Event()
#             self.clear_event = Event()
#             self.tsigma_process = TSigmaDeconstructor(self.build_new_learnblock,
#                                                       database,
#                                                       self.settings,
#                                                       self.tsigma_inputs,
#                                                       self.tsigma_outputs,
#                                                       self.upgrade_event,
#                                                       self.apply_event,
#                                                       self.clear_event)
#             self.tsigma_process.daemon = True
#             self.tsigma_process.start()
#
#         return self.tsigma_process
#
# #     def deconstruct_serial(self, level, pragmatic, block, database):
# #         changes_done_status = []
# #
# #         relatives = self.relative_finder.find(tuple(("complete", )),
# #             pragmatic, self.settings.deconst_full_tolerance)
# #
# #         deq_one = deque()
# #         decon_one = CompleteDeconstructor(level,
# #                                           database,
# #                                           self.settings,
# #                                           self.reconstructor,
# #                                           pragmatic,
# #                                           relatives,
# #                                           block,
# #                                           deq_one,
# #                                           cluster_finder=self.cluster_finder)
# #         decon_one.run()
# #         for log in deq_one:
# #             yield log
# #             changes_done_status.append(log.status)
# #             self.clean_ts_queue(log.deleted)
# #             if log.status and self.settings.deconst_mode == "minimal":
# #                 return
# #             elif pragmatic in log.deleted:
# #                 return
# #
# #         relatives = self.relative_finder.find(("T", "Sigma"),
# #             pragmatic, self.settings.deconst_full_tolerance)
# #         deq_two = deque()
# #         decon_two = TSigmaDeconstructor(self.build_new_learnblock,
# #                                         database,
# #                                         self.settings)
# #         decon_two.run_serial(level, relatives, pragmatic, block, deq_two)
# #         for log, level_, lb in deq_two:
# #             TS_QUEUE.append((level_, lb))
# #             yield log
# #             changes_done_status.append(log.status)
# #             if log.status and self.settings.deconst_mode == "minimal":
# #                 return
# #             elif pragmatic in log.deleted:
# #                 return
# #
# #         relatives = self.relative_finder.find(("Sigma", "Z"),
# #             pragmatic, self.settings.deconst_full_tolerance)
# #         deq_three = deque()
# #         decon_three = ZSigmaDeconstructor(level,
# #                                           database,
# #                                           self.settings,
# #                                           self.reconstructor,
# #                                           pragmatic,
# #                                           relatives,
# #                                           block,
# #                                           deq_three)
# #         decon_three.run()
# #         for log in deq_three:
# #             yield log
# #             self.clean_ts_queue(log.deleted)
# #             changes_done_status.append(log.status)
# #             if log.status and self.settings.deconst_mode == "minimal":
# #                 return
# #             elif pragmatic in log.deleted:
# #                 return
# #
# #         if not changes_done_status or not all(changes_done_status):
# #             database.insert(pragmatic)
# #             yield self.arrange_log(pragmatic)
#
#     def run_complete_deconstructor(self):
#         pass
#
#
#     def clean_ts_queue(self, deleted):
#         queue_copy = TS_QUEUE.copy()
#         for level, lb in queue_copy:
#             for model in deleted:
#                 if model in lb.origin:
#                     TS_QUEUE.remove((level, lb))
#
#     def arrange_log(self, model):
#         log = DeconstInfo()
#         log.pragmatic = model
#         log.inserted.append(model)
#         return log
