from dataclasses import dataclass
from multiprocessing import Process

from cml.domain.deconstruction.abstract import AbstractDeconstructor

__all__ = (
    "TSigmaDeconstructor",
    "Start",
    "Upgrade",
    "Apply",
    "Clean"
)


@dataclass
class Start:
    level: int
    relatives: list
    pragmatic: object


@dataclass
class Upgrade:
    deletes: set


@dataclass
class Apply:
    pass


@dataclass
class Clean:
    pass


class TSigmaDeconstructor(AbstractDeconstructor, Process):
    RELATION = ("Sigma", "T")

    def __init__(self, build_new_learnblock, knowledge, settings,
                 inputs=None, outputs=None,
                 upgrade_event=None, apply_event=None, clear_event=None):
        super(Process, self).__init__()

        self.build_new_learnblock = build_new_learnblock
        self.knowledge = knowledge
        self.settings = settings
        self.temp_changes = []

        # Used during parallelism
        self.inputs = inputs
        self.outputs = outputs
        self.upgrade_event = upgrade_event
        self.apply_event = apply_event
        self.clear_event = clear_event

    def run(self):
        while 1:
            command = self.inputs.get()

            if isinstance(command, Start):
                for relative in command.relatives:
                    log = self.arrange_logs(relative, command.pragmatic)
                    block = command.pragmatic.trained_with(self.knowledge)
                    self.deconstruct(command.level, relative, command.pragmatic,
                                     block, log)

            elif isinstance(command, Upgrade):
                self.upgrade(command.deletes)
                self.upgrade_event.set()

            elif isinstance(command, Apply):
                self.apply()
                self.apply_event.set()

            elif isinstance(command, Clean):
                self.clean()
                self.clear_event.set()

    def run_serial(self, level, relatives, pragmatic, result_deque):
        block = pragmatic.trained_with(self.knowledge)

        for relative in relatives:
            log = self.arrange_logs(relative, pragmatic)
            self.deconstruct(level, relative, pragmatic, block, log)

        for log, level, lb in self.temp_changes:
            result_deque.append((log, level, lb))
            if log.status and self.settings.deconst_mode == "minimal":
                self.temp_changes.clear()
                break

    def upgrade(self, deletions):
        to_delete = []
        for log, level, learnblock in self.temp_changes:
            if set(deletions).intersection(set(learnblock.origin)):
                to_delete.append((log, level, learnblock))

        for pair in to_delete:
            self.temp_changes.remove(pair)

    def apply(self):
        for log, level, lb in self.temp_changes:
            self.outputs.put((log, level, lb))
            if log.status:
                self.clean()
                return
        self.clean()

    def clean(self):
        self.temp_changes.clear()
        while not self.inputs.empty():
            self.inputs.get(timeout=0)

        while not self.inputs.empty():
            self.outputs.get(timeout=0)

    def deconstruct(self, level, relative, pragmatic, block, log):
        if self.level_constraint(level):
            overlapping = self.identify_overlapping_block(block, relative)
            if self.row_and_reliability_contraints(overlapping):
                block = self.prepare_higher_block(pragmatic, relative, overlapping)
                log.status = True
                log.new_learnblock = block
                self.temp_changes.append((log, level + 1, block))

    def level_constraint(self, level):
        return level < self.settings.highest_level - 1

    def identify_overlapping_block(self, block, relative):
        relative_block = relative.trained_with(self.knowledge)
        overlapped_block = block.get_overlapping(relative_block)
        return overlapped_block

    def row_and_reliability_contraints(self, block):
        return self.row_constraint(block) and self.reliability_constraint(block)

    def row_constraint(self, block):
        return len(block) >= self.settings.learn_block_minimum

    def reliability_constraint(self, block):
        alpha = self.calculate_reliability(
            block.get_column("Z"),
            block.get_column("Z_other"))
        alpha_systematic = alpha < 0
        alpha_weak_reliability = 0 <= alpha < self.settings.min_reliability
        return (self.settings.allow_weak_reliability and
                alpha_weak_reliability) or alpha_systematic

    def prepare_higher_block(self, pragmatic, relative, overlapping):

        # overblock = self.source.new_learnblock(
        overblock = self.build_new_learnblock(
            values=list(zip(
                overlapping.get_column("Z"),
                overlapping.get_column("Z_other"),
                overlapping.get_column("T"),
                ("\"\"" for _ in range(overlapping.n_rows())),
                ("\"\"" for _ in range(overlapping.n_rows())))),
            columns=(pragmatic.uid, relative.uid, "T", "Sigma", "Z"),
            index=[i for i in range(overlapping.n_rows())],
            origin=(pragmatic.uid, relative.uid))
        return overblock
